import django.dispatch


transaction_started = django.dispatch.Signal(providing_args=["transaction"])
transaction_succeeded = django.dispatch.Signal(providing_args=["transaction"])
transaction_failed = django.dispatch.Signal(providing_args=["transaction"])

auth_started = django.dispatch.Signal(providing_args=["auth"])
auth_succeeded = django.dispatch.Signal(providing_args=["auth"])
auth_failed = django.dispatch.Signal(providing_args=["auth"])



import os

from django.apps import apps
from django.conf import settings


def parse_banklinks(config=None):
    """Validates settings.BANKLINKS and add defaults to items

    for a working example see the definitions in `example/settings.py`
    """
    if config is None:
        config = getattr(settings, "BANKLINKS")

    result = {}

    if isinstance(config, dict):
        for bank_name, bank_data in config.items():
            if isinstance(bank_data, dict):
                final_data = {
                    "PROTOCOL": "ipizza",
                    "TYPE": "banklink",
                    "ORDER": 99,
                    "SEND_REF": True,
                    "PUBLIC_KEY": None,
                    "PRIVATE_KEY": None,
                }

                final_data.update(bank_data)

                if final_data["PUBLIC_KEY"] is not None and os.path.isfile(
                    final_data["PUBLIC_KEY"]
                ):
                    final_data["PUBLIC_KEY"] = os.path.abspath(final_data["PUBLIC_KEY"])

                if final_data["PRIVATE_KEY"] is not None and os.path.isfile(
                    final_data["PRIVATE_KEY"]
                ):
                    final_data["PRIVATE_KEY"] = os.path.abspath(
                        final_data["PRIVATE_KEY"]
                    )

                result[bank_name] = final_data

            else:
                result[bank_name] = bank_data

    return result


def get_model_name(model_name):
    manual = getattr(settings, "THORBANKS_MANUAL_MODELS", {})
    model_full_name = manual.get(model_name, None)

    return model_full_name


def get_model(model_name):
    """THORBANKS_MANUAL_MODELS can be used to make thorbanks work with your custom Authentication/Transaction model

    Example settings:
        # Note: Make sure that "thorbanks_models" is not in INSTALLED_APPS
        # Note 2: With manual models one does not need to add thorbanks under `MIGRATION_MODULES`

        THORBANKS_MANUAL_MODELS = {
            "Authentication": "myapp.Authentication",
            "Transaction": "myapp.Transaction",
        }

    Example myapp/models.py:
        # Note: If you only use Authentication then feel free to not add Transaction model (or vice-versa).

        class Authentication(AbstractAuthentication):
            pass

        class Transaction(AbstractTransaction):
            pass

    Finally run `makemigrations myapp` to create migrations for newly registered models.
    """
    model_full_name = get_model_name(model_name) or "thorbanks_models.%s" % model_name
    return apps.get_model(model_full_name)


# Shorthand methods
def get_private_key(the_bank):
    return get_links()[the_bank]["PRIVATE_KEY"]


def get_public_key(the_bank):
    """

    Note: To extract a public key out of a bank certificate use the following command:

        $ openssl x509 -pubkey -noout -in cert.pem  > pubkey.pem
    """
    return get_links()[the_bank]["PUBLIC_KEY"]


def get_client_id(the_bank):
    return get_links()[the_bank]["CLIENT_ID"]


def get_bank_id(the_bank):
    return get_links()[the_bank]["BANK_ID"]


def get_request_url(the_bank):
    return get_links()[the_bank]["REQUEST_URL"]


def get_link_type(the_bank):
    return get_links()[the_bank]["TYPE"]


def get_link_protocol(the_bank):
    return get_links()[the_bank]["PROTOCOL"]


def get_send_ref(the_bank):
    return get_links()[the_bank]["SEND_REF"]


def get_bank_choices():
    """Returns list of (bank_name, pretty_name, image_path, order) tuples.

    Useful in forms
    """
    links = get_links()

    return [
        (
            the_bank,
            links[the_bank].get("PRINTABLE_NAME", the_bank),
            links[the_bank].get("IMAGE_PATH", the_bank),
            links[the_bank].get("ORDER", 99),
        )
        for the_bank in links
    ]


# Updated by configure method below. Do not use directly and access this value trough get_links method
_LINKS = None


def get_links():
    global _LINKS

    if _LINKS is None:
        configure()

    return _LINKS


def configure(__only_use_during_tests=None):
    global _LINKS

    _LINKS = parse_banklinks()

    if isinstance(__only_use_during_tests, dict):
        for k, v in __only_use_during_tests.items():
            _LINKS[k].update(v)



from django import forms
from django.conf import settings as django_settings
from django.forms.widgets import RadioSelect
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.utils.encoding import force_str
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from thorbanks import settings
from thorbanks.settings import get_send_ref
from thorbanks.signals import transaction_started
from thorbanks.utils import calculate_731_checksum, create_signature


class AuthRequestBase(forms.Form):
    def __init__(
        self, bank_name, redirect_to, response_url, *args, extra_fields=None, **kwargs
    ):
        self.auth = settings.get_model("Authentication")()
        self.auth.bank_name = bank_name
        self.auth.redirect_after_success = redirect_to
        self.auth.redirect_on_failure = redirect_to

        if extra_fields is not None:
            for key, value in extra_fields.items():
                setattr(self.auth, key, value)

        self.auth.save()

        initial = self.prepare(bank_name, redirect_to, response_url, *args, **kwargs)

        super(AuthRequestBase, self).__init__(initial, *args, **kwargs)

        if not self.is_valid():
            raise RuntimeError("invalid initial data")  # pragma no cover

        self.finalize()
        if not self.is_valid():
            raise RuntimeError("signature is invalid")  # pragma no cover

    def prepare(self, bank_name, redirect_to, response_url, *args, **kwargs):
        raise NotImplementedError  # pragma no cover

    def finalize(self):
        raise NotImplementedError  # pragma no cover

    def redirect_html(self):
        """ Redirection html """
        html = (
            '<form action="%s" method="POST" id="banklink_redirect_url" accept-charset="%s">'
            % (self.get_request_url(), self.get_encoding())
        )

        for field in self:
            html += force_str(field) + "\n"

        html += "</form>"
        html += """<script type="text/javascript">
                       document.forms['banklink_redirect_url'].submit();
                   </script>
                """

        return mark_safe(html)

    def get_request_url(self):
        return settings.get_request_url(self.auth.bank_name)

    def get_encoding(self):
        return "UTF-8"

    def get_redirect_response_html(self):
        return render_to_string("thorbanks/auth-request.html", {"form": self})

    def get_redirect_response(self):
        return HttpResponse(self.get_redirect_response_html(), content_type="text/html")


class IPizzaAuthRequest(AuthRequestBase):
    VK_SERVICE = forms.CharField(widget=forms.HiddenInput())
    VK_VERSION = forms.CharField(widget=forms.HiddenInput())
    VK_SND_ID = forms.CharField(widget=forms.HiddenInput())
    VK_REC_ID = forms.CharField(widget=forms.HiddenInput())
    VK_NONCE = forms.CharField(widget=forms.HiddenInput(), max_length=50)
    VK_RETURN = forms.CharField(widget=forms.HiddenInput())
    VK_DATETIME = forms.CharField(widget=forms.HiddenInput())

    VK_ENCODING = forms.CharField(widget=forms.HiddenInput())

    VK_RID = forms.CharField(widget=forms.HiddenInput(), required=False)
    VK_MAC = forms.CharField(widget=forms.HiddenInput(), required=False)

    def prepare(self, bank_name, redirect_to, response_url, *args, **kwargs):
        initial = {
            "VK_SERVICE": "4012",
            "VK_VERSION": "008",
            "VK_SND_ID": settings.get_client_id(bank_name),
            "VK_REC_ID": settings.get_bank_id(bank_name),
            "VK_NONCE": self.auth.pk,
            "VK_RETURN": response_url,
            "VK_DATETIME": self.auth.created.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "VK_RID": "",
            "VK_REPLY": "3013",
            "VK_ENCODING": self.get_encoding(),
        }
        return initial

    def finalize(self):
        mac = create_signature(self.cleaned_data, self.auth.bank_name, auth=True)
        self.data["VK_MAC"] = mac


class PaymentRequestBase(forms.Form):
    def __init__(self, *args, extra_fields=None, **kwargs):
        self.transaction = settings.get_model("Transaction")()
        self.transaction.bank_name = kwargs["bank_name"]
        self.transaction.description = kwargs["message"]
        self.transaction.amount = round(kwargs["amount"], 2)
        self.transaction.currency = kwargs["currency"]
        self.transaction.message = kwargs["message"]

        self.transaction.redirect_after_success = kwargs["redirect_to"]
        self.transaction.redirect_on_failure = kwargs["redirect_on_failure"]

        if extra_fields is not None:
            for key, value in extra_fields.items():
                setattr(self.transaction, key, value)

        self.transaction.save()

        transaction_started.send(
            settings.get_model("Transaction"), transaction=self.transaction
        )

        initial = self.prepare(self.transaction, kwargs["url"])

        super(PaymentRequestBase, self).__init__(initial, *args)

        if not self.is_valid():
            raise RuntimeError("invalid initial data")

        self.finalize()
        if not self.is_valid():
            raise RuntimeError("signature is invalid")

    def prepare(self, transaction, url, language="EST"):
        raise NotImplementedError

    def finalize(self):
        raise NotImplementedError

    def redirect_html(self):
        html = (
            '<form action="%s" method="POST" id="banklink_redirect_url" accept-charset="%s">'
            % (self.get_request_url(), self.get_encoding())
        )
        for field in self:
            html += force_str(field) + "\n"
        html += "</form>"
        html += """<script type="text/javascript">
                    document.forms['banklink_redirect_url'].submit();
                    </script>"""
        return mark_safe(html)

    def submit_button(self, value="Make the payment"):
        html = '<form action="%s" method="POST">' % (
            settings.get_request_url(self.transaction.bank_name)
        )
        for field in self:
            html += force_str(field) + "\n"
        html += '<input type="submit" value="%s" />' % value
        html += "</form>"
        return mark_safe(html)

    def get_request_url(self):
        return settings.get_request_url(self.transaction.bank_name)

    def get_encoding(self):
        return "UTF-8"

    def get_redirect_response_html(self):
        return render_to_string("thorbanks/payment-request.html", {"form": self})

    def get_redirect_response(self):
        return HttpResponse(self.get_redirect_response_html(), content_type="text/html")


class PaymentRequest(PaymentRequestBase):
    """Creates payment request and Transaction object.

    It also acts as a facade through which transaction object and HTTP redirect can be retrieved.
    """

    VK_SERVICE = forms.CharField(widget=forms.HiddenInput())
    VK_VERSION = forms.CharField(widget=forms.HiddenInput())
    VK_SND_ID = forms.CharField(widget=forms.HiddenInput())
    VK_STAMP = forms.CharField(widget=forms.HiddenInput())
    VK_AMOUNT = forms.CharField(widget=forms.HiddenInput())
    VK_CURR = forms.CharField(widget=forms.HiddenInput())
    VK_REF = forms.CharField(widget=forms.HiddenInput(), required=False)
    VK_MSG = forms.CharField(widget=forms.HiddenInput())
    VK_RETURN = forms.CharField(widget=forms.HiddenInput())
    VK_CANCEL = forms.CharField(widget=forms.HiddenInput())
    VK_DATETIME = forms.CharField(widget=forms.HiddenInput())
    VK_MAC = forms.CharField(widget=forms.HiddenInput(), required=False)
    VK_LANG = forms.CharField(widget=forms.HiddenInput())
    VK_ENCODING = forms.CharField(widget=forms.HiddenInput())

    def prepare(self, transaction, url, language="EST"):
        assert language in ["EST", "ENG", "RUS"]

        return {
            "VK_SERVICE": "1012",
            "VK_VERSION": "008",
            "VK_ENCODING": self.get_encoding(),
            "VK_DATETIME": transaction.created.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "VK_RETURN": url,
            "VK_CANCEL": url,
            "VK_LANG": language,
            "VK_SND_ID": settings.get_client_id(transaction.bank_name),
            "VK_STAMP": transaction.pk,
            "VK_REF": calculate_731_checksum(transaction.pk)
            if get_send_ref(transaction.bank_name)
            else "",
            "VK_AMOUNT": transaction.amount,
            "VK_CURR": transaction.currency,
            "VK_MSG": transaction.message,
        }

    def finalize(self):
        self.data["VK_MAC"] = create_signature(
            self.cleaned_data, self.transaction.bank_name
        )


class PaymentFormMixin:
    """
    Base form mixin that lets the user choose the bank used for payment.
    Note that you MUST add bank_name field to your form class, you can use `get_bank_name_field()` method to get a field
    object for that. E.g.

    ```
    class OrderForm(PaymentFormMixin, forms.Form):
        bank_name = PaymentFormMixin.get_bank_name_field()
        # Add your custom fields and/or inherit from ModelForm instead of Form
    ```
    """

    class BankNameFieldWidget(RadioSelect):
        """Augmented RadioSelect that lets us use our own CSS class."""

        template_name = "thorbanks/payment_widget.html"
        option_template_name = "thorbanks/input_option.html"

    def __init__(self, *args, **kwargs):
        self.banklink_order_overwrite = kwargs.pop("banklink_order_overwrite", dict())
        # None means all banks in the settings are permitted
        payment_methods = kwargs.pop("payment_methods", None)

        super(PaymentFormMixin, self).__init__(*args, **kwargs)

        if not self.fields["bank_name"].required:
            raise NotImplementedError("bank_name field must be required")

        self.fields["bank_name"].choices = self.get_payment_method_choices(
            payment_methods
        )
        self.fields["bank_name"].error_messages["required"] = _(
            "You have to select a method of payment."
        )

    @classmethod
    def get_bank_name_field(cls):
        return forms.ChoiceField(
            label="Bank name", required=True, widget=cls.BankNameFieldWidget()
        )

    def get_payment_method_choices(self, all_payment_methods):
        payment_choices = []
        label_pattern = """<img alt="%s" src="%s" />"""

        for bank_id, bank_name, image_name, order in settings.get_bank_choices():
            if (
                self.banklink_order_overwrite
                and bank_id in self.banklink_order_overwrite
            ):
                order = self.banklink_order_overwrite[bank_id]

            if all_payment_methods is None or bank_id in all_payment_methods:
                label = label_pattern % (
                    bank_name,
                    self.get_bank_image_path(image_name),
                )
                payment_choices.append((bank_id, mark_safe(label), order))

        return [(y[0], y[1]) for y in sorted(payment_choices, key=lambda x: x[2])]

    def get_bank_image_path(self, image_name):
        logos_dir = getattr(django_settings, "BANKLINK_LOGO_PATH", "img/payment/")
        return "%s%s%s" % (django_settings.STATIC_URL, logos_dir, image_name)

    def get_bank_id(self):
        if not self.cleaned_data:
            return None

        return self.cleaned_data["bank_name"]



from django.db import models
from django.utils.translation import gettext_lazy as _


class AbstractTransaction(models.Model):
    STATUS_PENDING = "P"
    STATUS_FAILED = "F"
    STATUS_COMPLETED = "C"
    STATUS_CHOICES = (
        (STATUS_PENDING, _("pending")),
        (STATUS_FAILED, _("failed")),
        (STATUS_COMPLETED, _("completed")),
    )

    bank_name = models.CharField(_("bank"), max_length=16)
    description = models.CharField(
        _("reference description"), max_length=255, help_text=_("reference description")
    )
    amount = models.DecimalField(_("amount"), decimal_places=2, max_digits=9)
    currency = models.CharField(_("currency"), max_length=3)
    created = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    status = models.CharField(
        _("status"), max_length=1, default=STATUS_PENDING, choices=STATUS_CHOICES
    )
    redirect_after_success = models.CharField(max_length=255, editable=False)
    redirect_on_failure = models.CharField(max_length=255, editable=False)

    def __str__(self):
        return _("Transaction {} - {} {:.2f} from {} [{}]").format(
            self.id,
            self.currency,
            self.amount,
            self.bank_name,
            self.get_status_display(),
        )

    class Meta:
        abstract = True
        verbose_name = _("transaction")
        ordering = ["-last_modified"]


class AbstractAuthentication(models.Model):
    STATUS_PENDING = "P"
    STATUS_FAILED = "F"
    STATUS_COMPLETED = "C"
    STATUS_CHOICES = (
        (STATUS_PENDING, _("Pending")),
        (STATUS_FAILED, _("Failed")),
        (STATUS_COMPLETED, _("Completed")),
    )

    status = models.CharField(
        _("Status"), max_length=1, default=STATUS_PENDING, choices=STATUS_CHOICES
    )
    bank_name = models.CharField(_("Bank"), max_length=16)
    created = models.DateTimeField(auto_now_add=True)
    last_modified = models.DateTimeField(auto_now=True)
    redirect_after_success = models.CharField(max_length=255, editable=False)
    redirect_on_failure = models.CharField(max_length=255, editable=False)
    raw_response = models.TextField(blank=True)

    def __str__(self):
        return "Authentication %s - %s [%s]" % (
            self.id,
            self.bank_name,
            self.get_status_display(),
        )  # pragma no cover

    class Meta:
        abstract = True
        verbose_name = _("Authentication")
        ordering = ["-last_modified"]



import os

from django.conf import settings
from django.core.checks import Error, register

from thorbanks.settings import configure, parse_banklinks


@register
def check_model_settings(app_configs, **kwargs):
    issues = []

    manual_models = getattr(settings, "THORBANKS_MANUAL_MODELS", None)

    if manual_models is None:  # No manual models
        # If no manual models then we need to ensure that `thorbanks_models` is configured correctly
        if "thorbanks_models" not in settings.INSTALLED_APPS:
            issues.append(
                Error(
                    "thorbanks_models must be added to settings.INSTALLED_APPS when not using THORBANKS_MANUAL_MODELS",
                    id="thorbanks.E001",
                )
            )

        migration_modules = getattr(settings, "MIGRATION_MODULES", {})

        if not migration_modules.get("thorbanks_models", ""):
            issues.append(
                Error(
                    "Thorbanks is missing from settings.MIGRATION_MODULES",
                    hint="Add it to your settings like this - `MIGRATION_MODULES = "
                    '{ "thorbanks_models": "shop.thorbanks_migrations" }.',
                    id="thorbanks.E002",
                )
            )

    else:
        if manual_models is not None and not isinstance(manual_models, dict):
            issues.append(
                Error(
                    "settings.THORBANKS_MANUAL_MODELS must be a dict",
                    hint="See docstring of thorbanks.settings.get_model.",
                    id="thorbanks.E003",
                )
            )

        if "thorbanks_models" in settings.INSTALLED_APPS:
            issues.append(
                Error(
                    "thorbanks_models should not be added to "
                    "settings.INSTALLED_APPS when using THORBANKS_MANUAL_MODELS",
                    id="thorbanks.E011",
                )
            )

    return issues


@register
def check_banklink_settings(app_configs, **kwargs):
    issues = []

    links = parse_banklinks(getattr(settings, "BANKLINKS", None))

    if links and isinstance(links, dict):
        # Verify it contains valid data
        for bank_name, data in links.items():
            if len(bank_name) > 16:
                issues.append(
                    Error(
                        "settings.BANKLINKS keys are limited to 16 characters ({})".format(
                            bank_name
                        ),
                        hint="See docstring of thorbanks.settings.parse_banklinks.",
                        id="thorbanks.E005",
                    )
                )

            if not isinstance(data, dict):
                issues.append(
                    Error(
                        "settings.BANKLINKS['{}'] must be a dict with settings for the bank".format(
                            bank_name
                        ),
                        hint="See docstring of thorbanks.settings.parse_banklinks.",
                        id="thorbanks.E006",
                    )
                )

                continue

            required_keys = [
                "REQUEST_URL",
                "PRIVATE_KEY",
                "PUBLIC_KEY",
                "CLIENT_ID",
                "BANK_ID",
                "PROTOCOL",
                "PRINTABLE_NAME",
                "IMAGE_PATH",
                "TYPE",
                "ORDER",
            ]

            if data["PROTOCOL"] == "ipizza":
                for key in required_keys:
                    if key not in data or data[key] is None:
                        issues.append(
                            Error(
                                "settings.BANKLINKS['{}']: {} is required".format(
                                    bank_name, key
                                ),
                                hint="See docstring of thorbanks.settings.parse_banklinks.",
                                id="thorbanks.E007",
                            )
                        )

                if data["PUBLIC_KEY"] is not None and not os.path.isfile(
                    data["PUBLIC_KEY"]
                ):
                    issues.append(
                        Error(
                            "settings.BANKLINKS['{}']: PUBLIC_KEY file `{}` does not exist".format(
                                bank_name, data["PUBLIC_KEY"]
                            ),
                            hint="See docstring of thorbanks.settings.parse_banklinks.",
                            id="thorbanks.E008",
                        )
                    )

                if data["PRIVATE_KEY"] is not None and not os.path.isfile(
                    data["PRIVATE_KEY"]
                ):
                    issues.append(
                        Error(
                            "settings.BANKLINKS['{}']: PRIVATE_KEY file `{}` does not exist".format(
                                bank_name, data["PRIVATE_KEY"]
                            ),
                            hint="See docstring of thorbanks.settings.parse_banklinks.",
                            id="thorbanks.E009",
                        )
                    )

            else:
                issues.append(
                    Error(
                        "settings.BANKLINKS['{}']: PROTOCOL must be ipizza".format(
                            bank_name
                        ),
                        hint="See docstring of thorbanks.settings.parse_banklinks.",
                        id="thorbanks.E010",
                    )
                )

    else:
        issues.append(
            Error(
                "settings.BANKLINKS must be a dict",
                hint="See docstring of thorbanks.settings.parse_banklinks for reference.",
                id="thorbanks.E004",
            )
        )

    configure()

    return issues



import logging

from django.http import HttpResponse, HttpResponseRedirect, QueryDict
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from thorbanks import settings
from thorbanks.forms import IPizzaAuthRequest, PaymentRequest
from thorbanks.signals import (
    auth_failed,
    auth_succeeded,
    transaction_failed,
    transaction_succeeded,
)
from thorbanks.utils import verify_signature


logger = logging.getLogger(__name__)


def get_request_data(request):
    if request.method == "POST":
        return request.POST

    return request.GET


class PaymentError(Exception):
    """Generic error class."""


class AuthError(Exception):
    """Generic error class."""


@csrf_exempt
def response(request):
    data = get_request_data(request)
    if "VK_MAC" not in data:
        raise PaymentError("VK_MAC not in request")

    klass = settings.get_model("Transaction")
    transaction = get_object_or_404(klass, pk=data["VK_STAMP"])

    # Set proper encoding and get the data again (this time in correct encoding)
    request.encoding = "UTF-8"

    # We have to manually replace POST after changing encoding because of a Django bug (?)
    request.POST = QueryDict(request.body, encoding=request.encoding)
    data = get_request_data(request)

    signature_valid = verify_signature(
        data, transaction.bank_name, data["VK_MAC"], response=True
    )
    if not signature_valid:
        raise PaymentError("Invalid signature. ")

    if data["VK_SERVICE"] == "1111":
        if transaction.status == klass.STATUS_PENDING:
            # Mark purchase as complete
            transaction.status = klass.STATUS_COMPLETED
            transaction.save()
            transaction_succeeded.send(klass, transaction=transaction)
    elif data["VK_SERVICE"] == "1911":
        if transaction.status == klass.STATUS_PENDING:
            # Mark purchase as failed
            transaction.status = klass.STATUS_FAILED
            transaction.save()
            transaction_failed.send(klass, transaction=transaction)
    else:
        logging.critical(
            "thorbanks.views.response(): Got invalid VK_SERVICE code from bank: %s (transaction %s)",
            data["VK_SERVICE"],
            data["VK_STAMP"],
        )

        raise PaymentError("Bank sent confirmation with invalid VK_SERVICE!")

    if data["VK_AUTO"] == "Y":
        # This is automatic pingback from the bank - send simple 200 response as text/plain.
        return HttpResponse("request handled", content_type="text/plain")

    # This request is from the user after being redirected from the bank to our server. Redirect her further.
    if data["VK_SERVICE"] == "1111":
        url = transaction.redirect_after_success
    else:
        url = transaction.redirect_on_failure

    return HttpResponseRedirect(url)


def create_payment_request(
    bank_name,
    message,
    amount,
    currency,
    pingback_url,
    redirect_to,
    redirect_on_failure,
    extra_fields=None,
):
    return PaymentRequest(
        bank_name=bank_name,
        amount=amount,
        currency=currency,
        redirect_to=redirect_to,
        redirect_on_failure=redirect_on_failure,
        message=message,
        url=pingback_url,
        extra_fields=extra_fields,
    )


def create_auth_request(
    request, bank_name, response_url, redirect_to=None, extra_fields=None
):
    request_form_classes = {
        "ipizza": IPizzaAuthRequest,
    }
    form_cls = request_form_classes.get(settings.get_link_protocol(bank_name))
    if form_cls is None:
        return None

    return form_cls(
        bank_name=bank_name,
        response_url=response_url,
        redirect_to=redirect_to or response_url,
        extra_fields=extra_fields,
    )


class AuthResponseView(View):
    def __init__(self, **kwargs):
        self.data = None
        self.url = None
        self.auth = None

        super(AuthResponseView, self).__init__(**kwargs)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        self.data = get_request_data(request)

        if "VK_MAC" in self.data:
            self.handle_ipizza(request)
        else:
            raise AuthError("Invalid request (unknown protocol)")

        return super(AuthResponseView, self).dispatch(request, *args, **kwargs)

    def handle_ipizza(self, request):
        klass = settings.get_model("Authentication")
        auth = get_object_or_404(klass, pk=self.data.get("VK_NONCE", None))

        # Set proper encoding and get the data again (this time in correct encoding)
        request.encoding = "UTF-8"

        # We have to manually replace POST after changing encoding because of a Django bug (?)
        request.POST = QueryDict(request.body, encoding=request.encoding)
        self.data = get_request_data(request)

        signature_valid = verify_signature(
            self.data, auth.bank_name, self.data["VK_MAC"], auth=True, response=True
        )
        if not signature_valid:
            raise AuthError("Invalid signature. ")

        self.data = self.data.dict()

        if self.data["VK_SERVICE"] == "3013":
            if auth.status == klass.STATUS_PENDING:
                # Mark auth as complete
                auth.raw_response = self.data
                auth.status = klass.STATUS_COMPLETED
                auth.save()
                auth_succeeded.send(klass, auth=auth)

        else:
            if auth.status == klass.STATUS_PENDING:
                # Mark auth as failed
                auth.raw_response = self.data
                auth.status = klass.STATUS_FAILED
                auth.save()
                auth_failed.send(klass, auth=auth)

        # This request is from the user after being redirected from the bank to our server. Redirect her further.
        if self.data["VK_SERVICE"] == "3013":
            self.url = auth.redirect_after_success
        else:
            self.url = auth.redirect_on_failure

        self.auth = auth

        self.data["name"] = self.data.get("VK_USER_NAME")
        self.data["person_code"] = self.data.get("VK_USER_ID")

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(self.url)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)



from base64 import b64decode, b64encode
from functools import reduce

from django.urls import reverse
from django.utils.encoding import force_str

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding

from thorbanks import settings


IPIZZA_REQUEST_ORDER = {
    "3012": (
        "VK_SERVICE",
        "VK_VERSION",
        "VK_USER",
        "VK_DATETIME",
        "VK_SND_ID",
        "VK_REC_ID",
        "VK_USER_NAME",
        "VK_USER_ID",
        "VK_COUNTRY",
        "VK_OTHER",
        "VK_TOKEN",
        "VK_RID",
    ),
    "3013": (
        "VK_SERVICE",
        "VK_VERSION",
        "VK_DATETIME",
        "VK_SND_ID",
        "VK_REC_ID",
        "VK_NONCE",
        "VK_USER_NAME",
        "VK_USER_ID",
        "VK_COUNTRY",
        "VK_OTHER",
        "VK_TOKEN",
        "VK_RID",
    ),
    "4011": (
        "VK_SERVICE",
        "VK_VERSION",
        "VK_SND_ID",
        "VK_REPLY",
        "VK_RETURN",
        "VK_DATETIME",
        "VK_RID",
    ),
    "4012": (
        "VK_SERVICE",
        "VK_VERSION",
        "VK_SND_ID",
        "VK_REC_ID",
        "VK_NONCE",
        "VK_RETURN",
        "VK_DATETIME",
        "VK_RID",
    ),
}


def get_ordered_request(request, auth=False, response=False):
    def append_if_exists(target, source, the_value):
        if the_value in source:
            target.append(source[the_value])
        return target

    if not auth:
        if response:
            expected_values = (
                "VK_SERVICE",
                "VK_VERSION",
                "VK_SND_ID",
                "VK_REC_ID",
                "VK_STAMP",
                "VK_T_NO",
                "VK_AMOUNT",
                "VK_CURR",
                "VK_REC_ACC",
                "VK_REC_NAME",
                "VK_SND_ACC",
                "VK_SND_NAME",
                "VK_REF",
                "VK_MSG",
                "VK_T_DATETIME",
            )

        else:
            expected_values = (
                "VK_SERVICE",
                "VK_VERSION",
                "VK_SND_ID",
                "VK_STAMP",
                "VK_AMOUNT",
                "VK_CURR",
                "VK_REF",
                "VK_MSG",
                "VK_RETURN",
                "VK_CANCEL",
                "VK_DATETIME",
            )

    else:
        if response:
            expected_values = IPIZZA_REQUEST_ORDER["3013"]

        else:
            expected_values = IPIZZA_REQUEST_ORDER["4012"]

    ordered_request = []
    for value in expected_values:
        ordered_request = append_if_exists(ordered_request, request, value)
    return ordered_request


def request_digest(request, bank_name, auth=False, response=False):
    """
    return request digest in Banklink signature form (see docs for format)
    """
    request = get_ordered_request(request, auth=auth, response=response)
    digest = ""
    for value in request:
        value_len = len(value)
        digest += force_str(value_len).rjust(3, "0")
        digest += force_str(value)
    return digest.encode("UTF-8")


def get_pkey(bank_name):
    with open(settings.get_private_key(bank_name), "rb") as handle:
        private_key = serialization.load_pem_private_key(
            handle.read(), password=None, backend=default_backend()
        )

    return private_key


def create_signature(request, bank_name, auth=False):
    """
    sign BankLink request in dict format with private_key
    """
    digest = request_digest(request, bank_name, auth=auth)

    private_key = get_pkey(bank_name)
    signature = private_key.sign(digest, padding.PKCS1v15(), hashes.SHA1())

    return force_str(b64encode(signature))


def verify_signature(request, bank_name, signature, auth=False, response=False):
    """
    verify BankLink reply signature
    """
    signature = force_str(signature)

    with open(settings.get_public_key(bank_name), "rb") as handle:
        public_key = serialization.load_pem_public_key(handle.read(), default_backend())

    digest = request_digest(request, bank_name, auth=auth, response=response)

    try:
        public_key.verify(
            b64decode(signature), digest, padding.PKCS1v15(), hashes.SHA1()
        )

        return True

    except InvalidSignature:
        return False


def weight_generator():
    """Used for weight generation by calculate_731_checksum"""
    while True:
        yield 7
        yield 3
        yield 1


def calculate_731_checksum(number):
    # Check if number is integer, then cast to string
    number = str(int(number))[::-1]
    gen = weight_generator()
    weight_sum = reduce(lambda x, y: x + int(y) * next(gen), number, 0)
    checksum = (10 - weight_sum % 10) % 10
    return int(number[::-1] + str(checksum))


def pingback_url(request=None, base_url=None):
    assert request or base_url

    if request:
        base_url = "%s://%s" % (
            "https" if request.is_secure() else "http",
            request.META["HTTP_HOST"],
        )

    return "%s%s" % (base_url, reverse("thorbanks_response"))



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string as base_import_string


def is_model_registered(app_label, model_name):
    """Checks whether a given model is registered."""
    try:
        apps.get_registered_model(app_label, model_name)
    except LookupError:
        return False
    else:
        return True


def validate_import_string(dotted_path):
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)

    except ValueError:
        raise ImproperlyConfigured("%s doesn't look like a module path" % dotted_path)

    return module_path, class_name


def import_string(dotted_path):
    return base_import_string(dotted_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['thorbanks', 'thorbanks_models']

package_data = \
{'': ['*'], 'thorbanks': ['static/img/payment/*', 'templates/thorbanks/*']}

install_requires = \
['Django>=1.11.27', 'cryptography>=2']

setup_kwargs = {
    'name': 'django-thorbanks',
    'version': '0.6.1',
    'description': '`django-thorbanks` provides a Django application for Estonian banklinks (iPizza protocol).',
    'long_description': '# django-thorbanks\n\n[![Build Status](https://travis-ci.org/thorgate/django-thorbanks.svg?branch=master)](https://travis-ci.org/thorgate/django-thorbanks)\n[![Coverage Status](https://coveralls.io/repos/github/thorgate/django-thorbanks/badge.svg?branch=master)](https://coveralls.io/github/thorgate/django-thorbanks?branch=master)\n[![PyPI release](https://badge.fury.io/py/django-thorbanks.png)](https://badge.fury.io/py/django-thorbanks)\n\n\nDjango app for integrating Estonian banklinks into your project.\n\n## Features\n\nBank            | Protocol    | Authentication      | Payment\n--------------- | ----------- | ------------------- | -------\nSwedbank        | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\nSEB             | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\nDanske          | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\nLHV             | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\nKrediidipank    | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\nNordea          | iPizza      | :heavy_check_mark:  | :heavy_check_mark:\n\n## Usage\n\n### 1. Install it:\n\n**Pip:**\n\n```bash\npip install django-thorbanks\n```\n\n**Pipenv:**\n\n```bash\npipenv install django-thorbanks\n```\n\n**Poetry:**\n\n```bash\npoetry add django-thorbanks\n```\n\n### 2. Add to installed apps\n\n```python\nINSTALLED_APPS = (\n    # Add the following apps:\n    "thorbanks",\n    "thorbanks_models",\n)\n```\n\n### 3. Configure and create migrations:\n\n**With MANUAL_MODELS:**\n\n- Remove `"thorbanks_models"` from `INSTALLED_APPS`\n- follow instructions from [thorbanks.settings.get_model](./thorbanks/settings.py#L59).\n\n**With default models:**\n\nMake django aware that thorbanks migrations are in your local apps folder via settings.MIGRATION_MODULES:\n\n> Note: Replace `shop` with the name of an existing app in your project.\n\n```python\n# Tell django that thorbanks_models migrations are in shop app (in thorbanks_migrations module)\nMIGRATION_MODULES = {"thorbanks_models": "shop.thorbanks_migrations"}\n```\n\nNow run `makemigrations thorbanks_models` and `migrate` management commands to create and apply the migrations.\n\n### 4. Add settings.BANKLINKS\n\nFor a working example see the definitions in [example/settings.py](example/settings.py).\n\n> Note:\n>You will need a public and private key for each bank integration.\n>In case you don\'t have the public key, you can generate one out of a certificate by:\n>```\n>openssl x509 -pubkey -noout -in cert.pem  > pubkey.pem\n>```\n\n### 5. Link Transaction to your Order model\n\n> Note: When using MANUAL_MODELS replace `thorbanks_models` with your local app name\n\n```python\nclass Order(models.Model):\n    # ... other fields\n    transaction = models.OneToOneField(\n        "thorbanks_models.Transaction", null=True, on_delete=models.SET_NULL\n    )\n```\n\n### 6. Include thorbanks urls\n\n```python\nurlpatterns = [\n    # This is where the user will be redirected after returning from the banklink page\n    url(r"^banks/", include("thorbanks.urls")),\n]\n```\n\n### 7. Add listeners to banklinks success & failure callbacks:\n\nSee [example.shop.models.banklink_success_callback](example/shop/models.py#L23) and [example.shop.models.banklink_failed_callback](example/shop/models.py#L44).\n\n### 8. Create views and forms for payments:\n\nsee [example.shop.views](example/shop/views.py) and [example.shop.forms](example/shop/forms.py).\n\n## iPizza protocol\n\n- [Test service](https://banks.maximum.thorgate.eu/et/info)\n- [Swedbank](https://www.swedbank.ee/business/cash/ecommerce/ecommerce?language=EST)\n    - [Spec](https://www.swedbank.ee/static/pdf/business/d2d/paymentcollection/Pangalingi_paringute_tehniline_spetsifikatsioon_09_10_2014.pdf)\n- [SEB](https://www.seb.ee/ariklient/igapaevapangandus/maksete-kogumine/maksete-kogumine-internetis/pangalingi-tehniline)\n- [LHV Bank](https://www.lhv.ee/pangateenused/pangalink/)\n',
    'author': 'Thorgate',
    'author_email': 'info@thorgate.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://thorgate.eu',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
